%macro pam_gen_fs_get_data_sel(data_selection=);
/* Get data selection parts from data_sel domain object table. */
data _null_;
  set &pam_dolibref..data_sel;
  if data_sel_id eq "&data_selection";
  
  call symput('periodicityInterval',periodicity_interval_cd);
  call symput('periodicity',put(periodicity_no,8.));
  call symput('filteredSubsetTableName',data_sel_id);
  
  if data_sel_transpose_flg ne "" then do;
    call symput('transposeFlag',data_sel_transpose_flg);
  end;
  else do;
    call symput('transposeFlag',"BLANK");
  end;

  if kupcase(data_sel_interpolate_flg) eq "Y" then do;
    call symput('interpolationMethodTag',kupcase(interpolation_method_tag_cd));
  end;
  else do;
    call symput('interpolationMethodTag',"NONE");
  end;

  if kupcase(data_sel_aggregate_flg) eq "Y" then do;
    call symput('aggregationMethodTag',kupcase(aggregation_method_tag_cd));
    call symput('aggregationMethodEvent',kupcase(aggregation_method_event_cd));

    if periodicity_interval_cd ne "" and periodicity_no ne . then do;
      if kupcase(periodicity_interval_cd) eq "SECONDS" then do;
	    periodicitySeconds = periodicity_no;
	  end;
      else if kupcase(periodicity_interval_cd) eq "HOURS" then do;
        periodicitySeconds = periodicity_no * 3600;
	  end;
      else if kupcase(periodicity_interval_cd) eq "DAYS" then do;
        periodicitySeconds = periodicity_no * 86400;
	  end;
      call symput('periodicitySeconds', put(periodicitySeconds,8.));
    end;
  end;
  else do;
    call symput('aggregationMethodTag',"");
    call symput('aggregationMethodEvent',"");
    call symput('periodicitySeconds',"");
  end;

  if period_interval_cd ne "" and period_cnt ne 0 then do;
    call symput('dynamic',"Y");
    call symput('p_period_interval_cd',period_interval_cd);
    call symput('p_period_count',put(period_cnt,8.));
  end;
  else do;
    call symput('dynamic',"N");
    call symput('startDatetime',put(data_sel_start_dttm,datetime.));
    call symput('endDatetime',put(data_sel_end_dttm,datetime.));
    call symput('startDateTag',put(datepart(data_sel_start_dttm),date9.));
    call symput('endDateTag',put(datepart(data_sel_end_dttm),date9.));
    call symput('startDateEvent',put(datepart(data_sel_start_dttm),date9.));
    call symput('endDateEvent',put(datepart(data_sel_end_dttm),date9.));
  end;

  if data_sel_start_tm ne . and data_sel_end_tm ne . then do;
    call symput('startTimeSlice',put(data_sel_start_tm,time8.));
    call symput('endTimeSlice',put(data_sel_end_tm,time8.));
  end;
  else do;
    call symput('startTimeSlice',put('00:00:00't,time8.));
    call symput('endTimeSlice',put('23:59:59't,time8.));
  end;
run;
%if &dynamic eq Y %then %do;
  %pam_resolve_dynamic_dt(data_sel=&data_selection, sm_proj=);
%end;
%mend;

%macro pam_gen_fs_get_dim_table_names;
/* Get the table names for the Asset and Location dimension tables. */
%let assetDimTable = %pam_mfgcom_get_table_name(dimensionType=ASSET);
%let locDimTable = %pam_mfgcom_get_table_name(dimensionType=LOCATION);
%mend;

%macro pam_gen_fs_get_content(contentTypes=, dataSelection=);
%if %nrquote(&contentTypes) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs_get_content requires a value for parameter contentTypes;
  %let pam_gen_fs_rc = 8;
%end;
%if %nrquote(&dataSelection) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs_get_content requires a value for parameter dataSelection;
  %let pam_gen_fs_rc = 8;
%end;

/* Create a macro variable for each content type. */
%local numContentTypes;
%let numContentTypes = %eval(%sysfunc(countc(%nrbquote(&contentTypes), %str( )))+1); /*I18nOK:LINE*/

%do i=1 %to &numContentTypes&;
  %let contentType&i=%sysfunc(scan(&contentTypes, &i)); /*I18nOK:LINE*/
  %if &&contentType&i eq EVENTS %then %let longContentType&i = MAINEVENTSOURCE; 
  %if &&contentType&i eq TAGS %then %let longContentType&i = MAINSENSORSOURCE; 
%end;                                                    
 
/* Create a data set work.data_sel_<contentType> for each Content Type. */
%do i=1 %to &numContentTypes;
  proc sql noprint;
    create table work.data_sel_&&contentType&i as
      select * 
      from &pam_dolibref..data_sel_content
      where kupcase(content_type_cd) eq kupcase("&&longContentType&i") and
          data_sel_id eq "&dataSelection";
  quit;
%end;

%mend;

%macro pam_gen_fs_get_filter_content(filterContentTypes=, dataSelection=);
%if %nrquote(&filterContentTypes) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs requires a value for parameter filterContentTypes;
  %let pam_gen_fs_rc = 8;
%end;
%if %nrquote(&dataSelection) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs_get_filter_content requires a value for parameter dataSelection;
  %let pam_gen_fs_rc = 8;
%end;

/* Create a macro variable for each content type. */
%local numFilterContentTypes;
%let numFilterContentTypes = %eval(%sysfunc(countc(%nrbquote(&filterContentTypes), %str( )))+1); /*I18nOK:LINE*/

%do i=1 %to &numFilterContentTypes&;
  %let filterContentType&i=%sysfunc(scan(&filterContentTypes, &i)); /*I18nOK:LINE*/
  %if &&filterContentType&i eq EVENTS %then %let longFilterContentType&i = MAINEVENTSOURCE; 
  %if &&filterContentType&i eq TAGS %then %let longFilterContentType&i = MAINSENSORSOURCE;  
%end;                                                    
 
/* Create a data set work.data_sel_<filterContentType>_data_filter for each Filter Content Type. */
%do i=1 %to &numFilterContentTypes;

  proc sql noprint;
    create table work.data_sel_&&filterContentType&i.._data_filter as
      select * 
      from &pam_dolibref..data_sel_content_data_filter
      where kupcase(content_type_cd) eq kupcase("&&longFilterContentType&i") and
          data_sel_id eq "&dataSelection";
  quit;
%end;

%mend;

%macro pam_gen_fs_get_mart_meta_tag;
  %let tag_table_id = %pam_mfgcom_get_table_id(contentType = MAINSENSORSOURCE);
  %let tagDimTable = %pam_mfgcom_get_table_name(table_id = &tag_table_id);

  %local measureFactID;
  %let measureFactID = %pam_mfgcom_get_table_id(subjectType = TAGS);
  %let measureFactTable = %pam_mfgcom_get_table_name(table_id = &measureFactID);

  %if &p_table_names ne  %then %do;
    %let p_table_names = &p_table_names. &measureFactTable;
  %end;
  %else %do;
    %let p_table_names = &measureFactTable;
  %end;

  %let measure_val_col_nm = %pam_mfgcom_get_column_name(table_id = &measureFactID,
                                                        columnType = MEASURE_VALUE);
  %let measure_dttm_col_nm = %pam_mfgcom_get_column_name(table_id = &measureFactID,
                                                         columnType = MEASURE_DTTM);
  %let RKColumnName = %pam_mfgcom_get_column_name(table_id = &tag_table_id,
                                                  columnType = RETAINED_KEY);
  %let tagIdColumnName = %pam_mfgcom_get_column_name(table_id = &tag_table_id,
                                                     columnType = BUSINESS_KEY);
  %let tagTransposeColumnName = %pam_mfgcom_get_column_name(table_id = &tag_table_id,
                                                            columnType = TRANSPOSE_COLUMN);
%mend;

%macro pam_gen_fs_get_mart_meta_event;
  %let event_table_id = %pam_mfgcom_get_table_id(contentType = MAINEVENTSOURCE);
  %let eventDimTable = %pam_mfgcom_get_table_name(table_id = &event_table_id);

  %let assetDimTable = %pam_mfgcom_get_table_name(contentType = MAINASSETSOURCE);

  %local eventFactID;
  %let eventFactID = %pam_mfgcom_get_table_id(subjectType = EVENTS);
  %let eventFactTable = %pam_mfgcom_get_table_name(table_id = &eventFactID);

  %if &p_table_names ne  %then %do;
    %let p_table_names = &p_table_names. &eventFactTable;
  %end;
  %else %do;
    %let p_table_names = &eventFactTable;
  %end;

  %let event_value_col_nm = %pam_mfgcom_get_column_name(table_id = &eventFactID,
                                                        columnType = MEASURE_VALUE);
  %let event_dttm_col_nm = %pam_mfgcom_get_column_name(table_id = &eventFactID,
                                                       columnType = MEASURE_DTTM);
  %let eventRKColumnName = %pam_mfgcom_get_column_name(table_id = &event_table_id,
                                                       columnType = RETAINED_KEY);
  %let eventIdColumnName = %pam_mfgcom_get_column_name(table_id = &event_table_id,
                                                       columnType = BUSINESS_KEY);
  %let eventTransposeColumnName = %pam_mfgcom_get_column_name(table_id = &event_table_id,
                                                              columnType = TRANSPOSE_COLUMN);
  
%mend;

%macro pam_gen_fs_get_list_of_tags(data_selection=);
proc sql noprint;
  create table work.data_sel_tags as
  select distinct content_id as tag_id
  from &pam_dolibref..data_sel_content
  where kupcase(content_type_cd) eq "MAINSENSORSOURCE" and 
        kupcase(data_sel_id) eq kupcase("&data_selection");
quit;
proc sql noprint;
  create table work.tagIDs as
  select data_sel_fact.fact_tbl_cd, data_sel_tags.tag_id
  from work.data_sel_tags as data_sel_tags,
       &pam_dolibref..data_sel as data_sel,
       &pam_dolibref..data_sel_fact as data_sel_fact
  where data_sel_fact.data_sel_id eq data_sel.data_sel_id and
        data_sel.data_sel_id eq "&data_selection" and
        kupcase(data_sel_fact.fact_tbl_cd) eq "TAGS";
quit;

proc sql noprint;
  select count(distinct tag_id) into :ntags
  from work.tagIDs;
quit;

%let ntags=%cmpres(&ntags);

%do i=1 %to &ntags;
  %global tagList&i;
%end;

proc sql noprint;
  select distinct quote(ktrim(kleft(tag_id)))
  into :tagList1 - :tagList&ntags
  from work.tagIDs;
quit;

%global tagListLength;
%let tagListLength = %sysfunc(compress(&sqlobs));

proc sql noprint;
  create table tagIDsAndRKs as
  select distinct tagIDS.tag_id, &RKColumnName
  from work.tagIDs as tagIDS, 
       &pam_dimlibref..&tagDimTable as tagDim
  where tagIDs.tag_id eq tagDim.tag_id and current_num eq 1;
quit;
data tagIDToRKFmt(keep=fmtname start label type);
  set work.tagIDsAndRKs;
  retain fmtname "tagIDToRKfmt" type "C";
  start=tag_id;
  label=&RKColumnName;
run;
proc sort data=work.tagIDToRKFmt nodupkey;
  by start label;
run;
proc format cntlin=tagIDToRKFmt;
run;
%mend;

/* Gets the RKs for all tags in the list of selected tags. */
%macro pam_gen_fs_get_tag_RKs;
proc sql noprint;
  select count(distinct &RKColumnName) into :ntagRKs
  from &pam_dimlibref..&tagDimTable
  where &tagIdColumnName in (
     %do i=1 %to %eval(&tagListLength. -1);
        &&tagList&i,
     %end;
        &&tagList&tagListLength) 
     and current_num eq 1;
quit;

%let ntagRKs=%cmpres(&ntagRKs);

%do i=1 %to &ntagRKs;
  %global tagRKList&i;
%end;

proc sql noprint;
  select &RKColumnName into :tagRKList1 - :tagRKList&ntagRKs
  from &pam_dimlibref..&tagDimTable
  where &tagIdColumnName in (
     %do i=1 %to %eval(&tagListLength. -1);
        &&tagList&i,
     %end;
        &&tagList&tagListLength)
     and current_num eq 1;
quit;
%global tagRKListLength;
%let tagRKListLength = %sysfunc(compress(&sqlobs));
%mend;

/* Perform the data selection against the fact table. */
%macro pam_gen_fs_get_extra_tag_obs;

%if &engine eq SASIOIMP or &engine eq IMPALA %then %do;
  data work.tagstats;
    set &pam_factlibref..pam_fact_tag_measures_stats(where=(
       &RKColumnName in (%do i=1 %to %eval(&tagRKListLength. -1);
                           &&tagRKList&i,
                         %end;
                           &&tagRKList&tagRKListLength)));
  run;
  proc sort data = work.tagstats out = stats;
    by &RKColumnName tag_measure_dt;
  run;
%end;
%else %do;
  proc sort data = &pam_factqlibref..pam_fact_tag_measures_stats
            out = stats;
     where &RKColumnName in (%do i=1 %to %eval(&tagRKListLength. -1);
                                &&tagRKList&i,
                             %end;
                             &&tagRKList&tagRKListLength);  
    by &RKColumnName tag_measure_dt;
  run;
%end;

data _null_;
  set work.stats;
  by &RKColumnName tag_measure_dt;
  retain startdate foundStart 0;
  retain overallStartDate;

  tag_measure_date = tag_measure_dt;
   
  if _n_ eq 1 then do;
    startDate = input("&startDateTag",date9.);
    overallStartDate = "31Dec9999"d;
  end;
  
  prevdt=lag(tag_measure_dt);
  
  prev_date = tag_measure_dt;

  if tag_measure_date ge startDate and foundStart eq 0 then do;
    if prev_date ne . then startDate = prev_date;
    else startDate = tag_measure_date;
    foundStart = 1;
  end;
  if last.pam_tag_dim_rk and tag_measure_date lt startDate and foundStart eq 0 then do;
    startDate = tag_measure_date;
    foundStart = 1;
  end;
  if last.&RKColumnName or foundStart then do;
    overallStartDate = min(overallStartDate,startDate);
    foundStart=0;
    call symput('startDateTag',put(overallStartDate,date9.));
    startDate = input("&startDateTag",date9.);
  end;
run;

%pam_delete_tables(tagstats);
%pam_delete_tables(stats);

%mend;

%macro pam_gen_fs_get_measure_facts;
  proc sql noprint;
    create table work.AssetAndTagRKs as
    select &RKColumnName, pam_asset_dim_rk
    from &pam_dimlibref..&tagDimTable
    where &RKColumnName in (
    %do i=1 %to %eval(&tagListLength. -1);
      &&tagRKList&i,
    %end;
      &&tagRKList&tagRKListLength);
  quit;
  proc sql noprint;
    create table work.AssetIdsAndTagRKs as
    select &RKColumnName, pam_asset_id
    from work.AssetAndTagRKs as RKs,
         &pam_dimlibref..&assetDimTable as AssetDim
    where RKs.pam_asset_dim_rk eq AssetDim.pam_asset_dim_rk;
  quit;
  data tagToAssetIdFmt(keep=fmtname start label type hlo);
    set work.AssetIdsAndTagRKs end=eof;
    retain fmtname "tagToAssetIdFmt_&data_sel_id._Z" type "N";
    start=&RKColumnName;
    label=pam_asset_id;
    output;
    if (eof) then 
    do;
       hlo = "O";
       start = -999;
       label = "UNKNOWN";
       output;
    end;
  run;
  proc sort data=work.tagToAssetIdFmt;
    by start label;
  run;
  proc format library=&pam_fmtlibref cntlin=tagToAssetIdFmt;
  run;
  data tagToAssetRKFmt(keep=fmtname start label type hlo);
    set work.AssetAndTagRKs end = eof;
    retain fmtname "tagToAssetRKFmt_&data_sel_id._Z" type "N";
    start=&RKColumnName;
    label=pam_asset_dim_rk;
    output;
    if (eof) then 
    do;
       hlo = "O";
       start = -999;
       label = .;
       output;
    end;
  run;
  proc sort data=work.tagToAssetRKFmt;
    by start label;
  run;
  proc format library=&pam_fmtlibref cntlin=tagToAssetRKFmt;
  run;

  proc sql noprint;
    create table work.tagData as
    select 
      %if &engine ne SASIOIMP and &engine ne IMPALA %then %do;
           put(pam_loc_dim_rk,locfmt_&data_sel_id._Z.) as LOC_CD length=20, 
           put(&RKColumnName,tagToAssetIdFmt_&data_sel_id._Z.) as ASSET_ID length=65,
      %end;
           input(put(&RKColumnName,tagToAssetRKFmt_&data_sel_id._Z.),best8.) as PAM_ASSET_DIM_RK, 
           &RKColumnName, 
           pam_loc_dim_rk, 
           &measure_dttm_col_nm, 
           &measure_val_col_nm
    from &pam_factqlibref..&measureFactTable
    where &RKColumnName in (
    %do i=1 %to %eval(&tagRKListLength. -1);
      &&tagRKList&i,
    %end;
      &&tagRKList&tagRKListLength) 
    %if &engine eq HADOOP or &engine eq SASIOIMP or &engine eq IMPALA %then %do;
      and partition_key in (&partitionsTag) 
    %end;
      and pam_date_dim_rk between &startNumTag and &endNumTag
    %if &engine eq HADOOP %then %do;
      ;
  quit;
  proc sort data=work.tagData;
    by pam_asset_dim_rk loc_cd &measure_dttm_col_nm;
  quit;
    %end;
    %else %if &engine ne SASIOIMP and &engine ne IMPALA %then %do;
      order by pam_asset_dim_rk, loc_cd, &measure_dttm_col_nm;
  quit;
    %end;
    %else %do;
      ;
  quit;
    %end;

  %if not(&sqlobs) %then %do;
    proc sql noprint;
      create table work.tagData (
        pam_loc_dim_rk numeric,
        &RKColumnName numeric,
        pam_asset_dim_rk numeric,
        &measure_dttm_col_nm numeric,
        &measure_val_col_nm numeric,
        LOC_CD varchar(20),
        ASSET_ID varchar(65)
	  );
    quit;
    %put;
    %put WARNING:  The current data selection definition results in no selected ;
    %put WARNING:  tag data.  You must change your data selection definition or ;
    %put WARNING:  load data into your data mart and re-run the data selection ;
    %put WARNING:  to see tag data in the filtered subset. ;
    %put;
    %let pam_gen_fs_rc = 4;
  %end;
  %else %if &engine eq SASIOIMP or &engine eq IMPALA %then %do;
    proc sql noprint;
      create table work.tagDataTemp as
      select 
             put(pam_loc_dim_rk,locfmt_&data_sel_id._Z.) as LOC_CD length=20, 
             put(&RKColumnName,tagToAssetIdFmt_&data_sel_id._Z.) as ASSET_ID length=65,
             pam_loc_dim_rk,
             &RKColumnName,
             pam_asset_dim_rk,
             &measure_dttm_col_nm,
             &measure_val_col_nm
      from work.tagData;
    quit;
    %pam_delete_tables(tagData);
    proc datasets library=work noprint;
      change tagDataTemp=tagData;
    run;
    proc sort data=work.tagData;
      by pam_asset_dim_rk loc_cd &measure_dttm_col_nm;
    quit;
  %end;
%mend;

%macro pam_gen_fs_get_list_of_assets(data_selection=);
proc sql noprint;
  create table work.data_sel_assets as
  select distinct content_id as asset_id
  from &pam_dolibref..data_sel_content
  where kupcase(content_type_cd) eq "MAINASSETSOURCE" and 
        kupcase(data_sel_id) eq kupcase("&data_selection");
quit;
%if not(&sqlobs) %then %do;
  proc sql noprint;
    select count(distinct pam_asset_id) into :nassets
    from &pam_dimlibref..&assetDimTable;
  quit;

  %let nassets=%cmpres(&nassets);

  %do i=1 %to &nassets;
    %global assetList&i;
  %end;

  proc sql noprint;
    select distinct quote(ktrim(kleft(pam_asset_id)))
    into :assetList1 - :assetList&nassets
    from &pam_dimlibref..&assetDimTable;
  quit;

  %global assetListLength;
  %let assetListLength = %sysfunc(compress(&sqlobs));

  %put;
  %put WARNING:  The current data selection definition contains no selected ;
  %put WARNING:  asset IDs.  All asset IDs in the asset dimension table in  ;
  %put WARNING:  the data mart will be used. ;
  %put;
  %let pam_gen_fs_rc = 4;
%end;
%else %do;
  proc sql noprint;
    select count(distinct asset_id) into :nassets
    from work.data_sel_assets;
  quit;

  %let nassets=%cmpres(&nassets);

  %do i=1 %to &nassets;
    %global assetList&i;
  %end;
 
  proc sql noprint;
    select distinct quote(ktrim(kleft(asset_id)))
    into :assetList1 - :assetList&nassets
    from work.data_sel_assets;
  quit;
 
  %global assetListLength;
  %let assetListLength = %sysfunc(compress(&sqlobs));
%end;
%mend;

%macro pam_gen_fs_get_list_of_events(data_selection=);

proc sql noprint;
  create table work.data_sel_events as
  select distinct content_id as event_id
  from &pam_dolibref..data_sel_content
  where kupcase(content_type_cd) eq "MAINEVENTSOURCE" and 
        kupcase(data_sel_id) eq kupcase("&data_selection");
quit;

proc sql noprint;
  create table work.eventIDs as
  select data_sel_fact.fact_tbl_cd, data_sel_events.event_id
  from work.data_sel_events as data_sel_events,
       &pam_dolibref..data_sel as data_sel,
       &pam_dolibref..data_sel_fact as data_sel_fact
  where data_sel_fact.data_sel_id eq data_sel.data_sel_id and
        data_sel.data_sel_id eq "&data_selection" and
        kupcase(data_sel_fact.fact_tbl_cd) eq "EVENTS";
quit;

proc sql noprint;
  select count(distinct event_ID) into :nevents
  from work.eventIDs;
quit;

%let nevents=%cmpres(&nevents);

%do i=1 %to &nevents;
  %global eventList&i;
%end;

proc sql noprint;
  select distinct quote(ktrim(kleft(event_id)))
  into :eventList1 - :eventList&nevents 
  from work.eventIDs;
quit;

%global eventListLength;
%let eventListLength = %sysfunc(compress(&sqlobs));

proc sql noprint;
  select count(distinct &eventRKColumnName) into :neventRKs
  from &pam_dimlibref..&eventDimTable
  where &eventIdColumnName in (
     %do i=1 %to %eval(&eventListLength. -1);
        &&eventList&i,
     %end;
        &&eventList&eventListLength) 
     and current_num eq 1;
quit;

%let neventRKs=%cmpres(&neventRKs);

%do i=1 %to &neventRKs;
  %global eventRKList&i;
%end;

proc sql noprint;
  select &eventRKColumnName into :eventRKList1 - :eventRKList&neventRKs 
  from &pam_dimlibref..&eventDimTable
  where &eventIdColumnName in (
     %do i=1 %to %eval(&eventListLength. -1);
        &&eventList&i,
     %end;
        &&eventList&eventListLength) 
     and current_num eq 1;
quit;
%global eventRKListLength;
%let eventRKListLength = %sysfunc(compress(&sqlobs));

proc sql noprint;
  create table work.eventColumnNames as
  select eventIDs.event_id, pam_event_dim.&eventTransposeColumnName, pam_event_dim.&eventRKColumnName
  from work.eventIDs as eventIDs,
       &pam_dimlibref..&eventDimTable as pam_event_dim
  where eventIDs.event_id eq pam_event_dim.event_id and current_num eq 1;
quit;

data eventIDToRKFmt(keep=fmtname start label type);
  set work.eventColumnNames;
  retain fmtname "eventIDToRKfmt" type "C";
  start=event_id;
  label=&eventRKColumnName;
run;
proc sort data=work.eventIDToRKFmt;
  by start label;
run;
proc format cntlin=eventIDToRKFmt;
run;

%mend;

/* Gets the RKs for all assets in the list of selected assets. */
%macro pam_gen_fs_get_asset_RKs;
proc sql noprint;
  select count(distinct pam_asset_dim_rk) into :nassetRKs
  from &pam_dimlibref..&assetDimTable
  where current_num eq 1;
quit;

%let nassetRKs=%cmpres(&nassetRKs);

%do i=1 %to &nassetRKs;
  %global assetRKList&i;
%end;

proc sql noprint;
  select distinct pam_asset_dim_rk
  into :assetRKList1 - :assetRKList&nassetRKs
  from &pam_dimlibref..&assetDimTable
  where
    %if &assetList1 ne %str() %then %do;
      pam_asset_id in (
        %do i=1 %to %eval(&assetListLength. -1);
          &&assetList&i,
        %end;
          &&assetList&assetListLength) and
    %end;
    current_num eq 1;
quit;

%global assetRKListLength;
%let assetRKListLength = %sysfunc(compress(&sqlobs));

%mend;

%macro pam_gen_fs_get_event_facts;
  proc sql noprint;
    create table work.assetIDs as
    select distinct pam_asset_dim_rk, pam_asset_ID
    from &pam_dimlibref..&assetDimTable
    where pam_asset_dim_rk in (
       %do i=1 %to %eval(&assetRKListLength. -1);
         &&assetRKList&i,
       %end;
         &&assetRKList&assetRKListLength)
       and current_num=1;
  quit;
  data assetRKfmt(keep=start label fmtname type hlo);
    set work.assetIDs(rename=(pam_asset_dim_rk=start pam_asset_id=label)) end=eof;
    retain fmtname "assetRKfmt_&data_sel_id._Z" type "N";
    output;
    if (eof) then do;
      hlo = "O";
      start = -999;
      label = "UNKNOWN";
      output;
    end;
  run;
  proc sort data=work.assetRKfmt;
    by start label;
  run;
  proc format library=&pam_fmtlibref cntlin=work.assetRKfmt;
  run;

  proc sql noprint;
    create table work.eventData as
    select 
      %if &engine ne SASIOIMP and &engine ne IMPALA %then %do;
           put(pam_loc_dim_rk,locfmt_&data_sel_id._Z.) as LOC_CD length=20, 
           put(pam_asset_dim_rk,assetRKfmt_&data_sel_id._Z.) as ASSET_ID length=65,
      %end;
           &eventRKColumnName, 
           pam_asset_dim_rk, 
           pam_loc_dim_rk,
           &event_dttm_col_nm, 
           &event_value_col_nm
    from &pam_factqlibref..&eventFactTable
    where &eventRKColumnName in (
       %do i=1 %to %eval(&eventRKListLength. -1);
         &&eventRKList&i,
       %end;
         &&eventRKList&eventRKListLength) and
      %if &assetRKList1 ne %str() %then %do;
          pam_asset_dim_rk in (
             %do i=1 %to %eval(&assetRKListLength. -1);
                &&assetRKList&i,
             %end;
                &&assetRKList&assetRKListLength) and
      %end;
      %if &engine eq HADOOP or &engine eq SASIOIMP or &engine eq IMPALA %then %do;
          partition_key in (&partitionsEvent) and
      %end;
          pam_date_dim_rk between &startNumEvent and &endNumEvent
      %if &engine eq HADOOP %then %do;
          ;
        quit;
        proc sort data=work.eventData;
          by pam_asset_dim_rk pam_loc_dim_rk &event_dttm_col_nm;
        quit;
      %end;
      %else %if &engine ne SASIOIMP and &engine ne IMPALA %then %do;
          order by pam_asset_dim_rk, pam_loc_dim_rk, &event_dttm_col_nm;
        quit;
      %end;
      %else %do;
          ;
        quit;
      %end;

  %if not(&sqlobs) %then %do;
    proc sql noprint;
      create table work.eventData (
        pam_loc_dim_rk numeric,
        &eventRKColumnName numeric,
        pam_asset_dim_rk numeric,
        &event_dttm_col_nm numeric,
        &event_value_col_nm numeric,
        LOC_CD varchar(20),
        ASSET_ID varchar(65)
      );
    quit;

    %put;
    %put WARNING:  The current data selection definition results in no selected ;
    %put WARNING:  event data.  You must change your data selection definition or ;
    %put WARNING:  load data into your data mart and re-run the data selection ;
    %put WARNING:  to see event data in the filtered subset. ;
    %put;
    %let pam_gen_fs_rc = 4;
  %end;
  %else %if &engine eq SASIOIMP or &engine eq IMPALA %then %do;
  proc sql noprint;
    create table work.eventDataTemp as
    select 
           put(pam_loc_dim_rk,locfmt_&data_sel_id._Z.) as LOC_CD length=20, 
           put(pam_asset_dim_rk,assetRKfmt_&data_sel_id._Z.) as ASSET_ID length=65,
           &eventRKColumnName, 
           pam_asset_dim_rk, 
           pam_loc_dim_rk,
           &event_dttm_col_nm, 
           &event_value_col_nm
    from work.eventData;
  quit;
  %pam_delete_tables(eventData);
  proc datasets library=work noprint;
    change eventDataTemp=eventData;
  run;
  proc sort data=work.eventData;
    by pam_asset_dim_rk pam_loc_dim_rk &event_dttm_col_nm;
  quit;  
  %end;
%mend;

%macro pam_gen_fs_transpose(data_selection=);
%local tagRKsNoFilter1;
%local eventRKsNoFilter1;
%local tagFilter;
%local eventFilter;

%if &extractMeasures eq Y %then %do;
  proc sql noprint;
    create table work.data_sel_tags_data_filter as
    select data_sel_id, content_type_cd, content_id as tag_id, filter_order_no, filter_value_no, 
           filter_operator_cd, filter_logical_operator_cd
    from &pam_dolibref..data_sel_content_data_filter
    where data_sel_id eq "&data_selection" and kupcase(content_type_cd) eq "MAINSENSORSOURCE";
  quit;

  proc sort data=work.data_sel_tags_data_filter(where=(data_sel_id eq "&data_selection"))
          out=sortedFilters;
    by tag_id filter_order_no;
  run;
  
  %local dsid;
  %let dsid=%sysfunc(open(work.sortedFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;     /* There is at least one tag filter. */
    /* Create table containing tags which have data value filters. */
    proc sql noprint;
      create table filterTagRKs as
      select distinct put(tag_id,$tagIDToRKFmt.) as &RKColumnName
      from work.sortedFilters;
    quit;
    /* Create table containing tags with no data value filters. */
    proc sql noprint;
      create table tagRKsNoFilter as
      select distinct tagIDsAndRKs.&RKColumnName
      from work.tagIDsAndRKs as allTags
           where allTags.tag_id not in 
		   (select tag_id from work.sortedFilters);
    quit;
  %end;
  %else %do;  /* There are no tags with filters. */
    /* Create empty table since no tags have filters assigned. */
    data filterTagRKs;
      length &RKColumnName 8;
      stop;
    run;
    /* Create table containing tags with no data value filters. */
    proc sql noprint;
      create table tagRKsNoFilter as
      select distinct &RKColumnName
      from work.tagIDsAndRKs;
    quit;
  %end;

  %local dsid;
  %let dsid=%sysfunc(open(work.sortedFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;
    data _null_;
      set work.sortedFilters end=last;
      by tag_id filter_order_no;

      length tempFilter $32767;
      retain tempFilter;

      if first.tag_id then do;
        tempFilter = "( &RKcolumnName eq " || put(tag_id,$tagIDToRKFmt.) || " and ( ";
	    count+1;
      end;
      else do;
        tempFilter = ktrim(tempFilter) || " " || ktrim(kleft(filter_logical_operator_cd));
      end;
	  %if &transposeFlag eq Y %then %do;
        tempFilter = ktrim(tempFilter) || " &measure_val_col_nm " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
      %end;
	  %else %do;
        tempFilter = ktrim(tempFilter) || " value " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
	  %end;
      if last.tag_id then do;
        tempFilter = ktrim(tempFilter) || " ) ";
        if not last then tempFilter = ktrim(tempFilter) || " ) OR ";
        else tempFilter = ktrim(tempFilter) || " )";
        call symput('phrase'||left(put(count,8.)),ktrim(kleft(tempFilter)));
        call symput('nphrases',left(put(count,8.)));
      end;
    run;

    %let tagFilter=;
    %do i = 1 %to &nphrases;
      %let tagFilter=&tagFilter &&phrase&i;
    %end;
  %end;
%end;

%if &extractEvents eq Y %then %do;
  proc sql noprint;
    create table work.data_sel_events_data_filter as
    select data_sel_id, content_type_cd, content_id as event_id, filter_order_no, filter_value_no, 
           filter_operator_cd, filter_logical_operator_cd
    from &pam_dolibref..data_sel_content_data_filter
    where data_sel_id eq "&data_selection" and kupcase(content_type_cd) eq "MAINEVENTSOURCE";
  run;

  proc sort data=work.data_sel_events_data_filter(where=(data_sel_id eq "&data_selection"))
          out=sortedEventFilters;
    by event_id filter_order_no;
  run;
  
  %local dsid;
  %let dsid=%sysfunc(open(work.sortedEventFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedEventFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;     /* There is at least one tag filter. */
    /* Create table containing events which have data value filters. */
    proc sql noprint;
      create table filterEventRKs as
      select distinct put(event_id,$eventIDToRKFmt.) as &eventRKColumnName 
      from work.sortedEventFilters;
    quit;
    /* Create table containing events with no data value filters. */
    proc sql noprint;
      create table eventRKsNoFilter as
      select distinct &eventRKColumnName 
      from work.eventColumnNames
      where event_id not in (select event_id from work.sortedEventFilters);
    quit;
  %end;
  %else %do; /* There are no event filters. */
    /* Create empty table since no events have filters assigned. */
    data filterEventRKs;
      length &eventRKColumnName 8;
      stop;
    run;
	/* Create table containing events with no data value filters. */
    proc sql noprint;
      create table eventRKsNoFilter as
      select distinct &eventRKColumnName
      from work.eventColumnNames;
    quit;
  %end;


  %local dsid;
  %let dsid=%sysfunc(open(work.sortedEventFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedEventFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;
    data _null_;
      set work.sortedEventFilters end=last;
      by event_id filter_order_no;

      length tempFilter $32767;
      retain tempFilter;

      if first.event_id then do;
        tempFilter = "( &eventRKColumnName eq " || put(event_id,$eventIDtoRKfmt.) || " and ( ";
	    count+1;
      end;
	  else do;
        tempFilter = ktrim(tempFilter) || " " || ktrim(kleft(filter_logical_operator_cd));
      end;
      %if &transposeFlag eq Y %then %do;
        tempFilter = ktrim(tempFilter) || " &event_value_col_nm " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
      %end;
	  %else %do;
        tempFilter = ktrim(tempFilter) || " value " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
      %end;
      if last.event_id then do;
        tempFilter = ktrim(tempFilter) || " ) ";
        if not last then tempFilter = ktrim(tempFilter) || " ) OR ";
        else tempFilter = ktrim(tempFilter) || " )";
        call symput('phrase'||left(put(count,8.)),kleft(ktrim(tempFilter)));
        call symput('nphrases',left(put(count,8.)));
      end;
    run;

    %let eventFilter=;
    %do i = 1 %to &nphrases;
      %let eventFilter=&eventFilter &&phrase&i;
    %end;
  %end;
%end;

%if %sysfunc(ktrim(%sysfunc(kleft(&transposeFlag)))) eq Y %then %do;
  %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y %then %do;
    proc sql noprint;
      create table work.tagKeys as
      select distinct pam_tag_dim.&RKColumnName
      from work.tagIDs as tagIDs,
           &pam_dimlibref..&tagDimTable as pam_tag_dim
      where kupcase(tagIDs.tag_id) eq kupcase(pam_tag_dim.tag_id) and 
            tagIDs.tag_id eq pam_tag_dim.&tagIdColumnName and current_num eq 1;
    quit;
  %end;
  %if %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
    proc sql noprint;
      create table work.eventKeys as
      select distinct pam_event_dim.&eventRKColumnName
      from work.eventIDs as eventIDs,
           &pam_dimlibref..&eventDimTable as pam_event_dim
      where kupcase(eventIDs.event_id) eq kupcase(pam_event_dim.event_id) and
            eventIDs.event_id eq pam_event_dim.&eventIDColumnName and current_num eq 1;
    quit;
  %end;

  %pam_expand_transpose( outds=&pam_fslibref..&filteredSubsetTableName,
                         groupvar=pam_asset_dim_rk pam_loc_dim_rk, 
                         begindttm=&startDatetime,
                         enddttm=&endDatetime, 
                         timeslicebegin=&startTimeSlice,
                         timesliceend=&endTimeSlice, 
                         periodicity=&periodicitySeconds
              %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y %then %do;
			             ,
			             measure_inds=tagData, 
                         measure_dttmvar=&measure_dttm_col_nm, 
                         measure_rkvar=&RKColumnName,  
                         measure_idvar=tag_ID,
                         measure_var=&measure_val_col_nm,
                         measure_colnamevar=&tagTransposeColumnName, 
                         measure_dim=&pam_dimlibref..&tagDimTable,
                         measure_rklist=work.tagKeys, 
                         measure_int=&interpolationMethodTag,
                         measure_agg=&aggregationMethodTag,
                         measure_filter=&tagFilter,
                         measure_filterRKs=work.filterTagRKs, 
                         measure_RKsNoFilter=work.tagRKsNoFilter
			  %end;
              %if %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
			             ,
                         event_inds=eventData, 
                         event_dttmvar=&event_dttm_col_nm, 
                         event_rkvar=&eventRKColumnName,  
                         event_idvar=event_ID,
                         event_var=&event_value_col_nm,
                         event_colnamevar=&eventTransposeColumnName, 
                         event_dim=&pam_dimlibref..&eventDimTable,
                         event_rklist=work.eventKeys,
                         event_agg=&aggregationMethodEvent,
                         event_filter=&eventFilter,
                         event_filterRKs=work.filterEventRKs, 
                         event_RKsNoFilter=work.eventRKsNoFilter
               %end;
                         );

%end;
%else %if %sysfunc(ktrim(%sysfunc(kleft(&transposeFlag)))) eq N or
          %sysfunc(ktrim(%sysfunc(kleft(&transposeFlag)))) eq BLANK %then %do;
  %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y and
      %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
    proc sql noprint;	  
      create table tagColumnNames as
      select &RKColumnName, tag_id
      from &pam_dimlibref..&tagDimTable
      where &RKColumnName in (
        %do i=1 %to %eval(&tagRKListLength. -1);
          &&tagRKList&i,
        %end;
          &&tagRKList&tagRKListLength);
    quit;
    data tagRKfmt(keep=start label fmtname type);
      set work.tagColumnNames(rename=(&RKColumnName=start tag_id=label));
      retain fmtname "tagRKfmt" type "N";
    run;
    proc sort data=work.tagRKfmt;
      by start label;
    run;
    proc format cntlin=work.tagRKfmt;
    run;
    proc sql noprint;
      create table work.eventIDs as
      select distinct &eventRKColumnName, event_ID
      from &pam_dimlibref..&eventDimTable
      where &eventRKColumnName in (
         %do i=1 %to %eval(&eventRKListLength. -1);
           &&eventRKList&i,
         %end;
           &&eventRKList&eventRKListLength) 
         and current_num=1;
    quit;
    data eventRKfmt(keep=start label fmtname type);
      set work.eventIDs(rename=(&eventRKColumnName=start event_id=label));
      retain fmtname "eventRKfmt" type "N";
    run;
    proc sort data=work.eventRKfmt;
      by start label;
    run;
    proc format cntlin=work.eventRKfmt;
    run;
    proc sql noprint;
      create table work.assetIDs as
      select distinct pam_asset_dim_rk, pam_asset_ID
      from &pam_dimlibref..&assetDimTable
      where
         %if &assetRKList1 ne %str() %then %do;
           pam_asset_dim_rk in (
              %do i=1 %to %eval(&assetRKListLength. -1);
                &&assetRKList&i,
              %end;
                &&assetRKList&assetRKListLength) and
         %end;
         current_num=1;
    quit;
    data assetRKfmt(keep=start label fmtname type hlo);
      set work.assetIDs(rename=(pam_asset_dim_rk=start pam_asset_id=label)) end=eof;
      retain fmtname "assetRKfmt" type "N";
      output;
      if (eof) then do;
        hlo = "O";
        start = -999;
        label = "UNKNOWN";
        output;
      end;
    run;
    proc sort data=work.assetRKfmt;
      by start label;
    run;
    proc format cntlin=work.assetRKfmt;
    run;

	%let tagRKsNoFilter1 = ;
    data _null_;
      set work.tagRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numTagRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numTagRKsNoFilter',num);
      end;
    run;

	%let eventRKsNoFilter1 = ;
    data _null_;
      set work.eventRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numEventRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numEventRKsNoFilter',num);
      end;
    run;

    data work.tagData(keep=loc_cd asset_id datetime id value);
      set work.tagData (rename=(&measure_dttm_col_nm=DATETIME &measure_val_col_nm=VALUE)
                        where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        ));
      length ID $100;
      attrib DATETIME label="DATETIME";
      attrib VALUE label = "VALUE";

      ID = put(&RKColumnName,tagRKfmt.);
	  
      %if &tagFilter ne %str() %then %do;
        if &tagFilter
          %if &tagRKsNoFilter1 ne %str() %then %do;
             or (&RKColumnName in (
               %do i = 1 %to &numTagRKsNoFilter;
                 &&tagRKsNoFilter&i
               %end;
             ))
          %end;
        then output;
      %end;

    run;
    data work.eventData(keep=loc_cd asset_id datetime id value);
      set work.eventdata (rename=(&event_dttm_col_nm=DATETIME &event_value_col_nm=VALUE)
                          where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        ));
      ID = put(&eventRKColumnName,eventRKfmt.);
	  
      %if &eventFilter ne %str() %then %do;
        if &eventFilter
          %if &eventRKsNoFilter1 ne %str() %then %do;
             or (&eventRKColumnName in (
               %do i = 1 %to &numEventRKsNoFilter;
                 &&eventRKsNoFilter&i
               %end;
             ))
           %end;
        then output;
      %end;

    run;
    data &pam_fslibref..&filteredSubsetTableName(keep=loc_cd asset_id datetime id value);
      set work.tagData work.eventData;
    run;
    proc sort data=&pam_fslibref..&filteredSubsetTableName;
      by loc_cd asset_id datetime;
    run;
  %end;
  %else %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y %then %do;
    proc sql noprint;
	  create table tagColumnNames as
	  select &RKColumnName, tag_id
	  from &pam_dimlibref..&tagDimTable
	  where &RKColumnName in (
         %do i=1 %to %eval(&tagListLength. -1);
           &&tagRKList&i,
         %end;
           &&tagRKList&tagRKListLength);
	quit;
    data tagRKfmt(keep=start label fmtname type);
      set work.tagColumnNames(rename=(&RKColumnName=start tag_id=label));
      retain fmtname "tagRKfmt" type "N";
    run;
    proc sort data=work.tagRKfmt;
      by start label;
    run;
    proc format cntlin=work.tagRKfmt;
    run;

    data _null_;
      set work.tagRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numTagRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numTagRKsNoFilter',num);
      end;
    run;

    data &pam_fslibref..&filteredSubsetTableName(keep=loc_cd asset_id datetime id value);
      set work.tagData (rename=(&measure_dttm_col_nm=DATETIME &measure_val_col_nm=VALUE)
                        where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        ));

      length ID $100;
      attrib DATETIME label="DATETIME";
      attrib VALUE label = "VALUE";

      ID = put(&RKColumnName,tagRKfmt.);

      %if &tagFilter ne %str() %then %do;
        if &tagFilter
          %if &tagRKsNoFilter1 ne %str() %then %do;
             or (&RKColumnName in (
               %do i = 1 %to &numTagRKsNoFilter;
                 &&tagRKsNoFilter&i
               %end;
             ))
          %end;

        then output;
      %end;

    run;
    proc sort data=&pam_fslibref..&filteredSubsetTableName;
      by loc_cd asset_id datetime;
    run;

  %end;
  %else %if %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
    proc sql noprint;
      create table work.eventIDs as
      select distinct &eventRKColumnName, event_ID
      from &pam_dimlibref..&eventDimTable
      where &eventRKColumnName in (
         %do i=1 %to %eval(&eventRKListLength. -1);
           &&eventRKList&i,
         %end;
           &&eventRKList&eventRKListLength) 
         and current_num=1;
    quit;
    data eventRKfmt(keep=start label fmtname type);
      set work.eventIDs(rename=(&eventRKColumnName=start event_id=label));
      retain fmtname "eventRKfmt" type "N";
    run;
    proc sort data=work.eventRKfmt;
      by start label;
    run;
    proc format cntlin=work.eventRKfmt;
    run;
    proc sql noprint;
      create table work.assetIDs as
      select distinct pam_asset_dim_rk, pam_asset_ID
      from &pam_dimlibref..&assetDimTable
      where pam_asset_dim_rk in (
         %do i=1 %to %eval(&assetRKListLength. -1);
           &&assetRKList&i,
         %end;
           &&assetRKList&assetRKListLength) 
         and current_num=1;
    quit;
    data assetRKfmt(keep=start label fmtname type hlo);
      set work.assetIDs(rename=(pam_asset_dim_rk=start pam_asset_id=label)) end=eof;
      retain fmtname "assetRKfmt" type "N";
      output;
      if (eof) then do;
        hlo = "O";
        start = -999;
        label = "UNKNOWN";
        output;
      end;
    run;
    proc sort data=work.assetRKfmt;
      by start label;
    run;
    proc format cntlin=work.assetRKfmt;
    run;

	%let eventRKsNoFilter1 = ;
    data _null_;
      set work.eventRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numEventRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numEventRKsNoFilter',num);
      end;
    run;

    data &pam_fslibref..&filteredSubsetTableName(keep=loc_cd asset_id id datetime value);
      set work.eventData (rename=(&event_dttm_col_nm=DATETIME &event_value_col_nm=VALUE)
                          where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        ));

      length ID $100;
      attrib DATETIME label="DATETIME";
      attrib VALUE label = "VALUE";

      ASSET_ID = put(pam_asset_dim_rk,assetRKfmt.);
      ID = put(&eventRKColumnName,eventRKfmt.);

      %if &eventFilter ne %str() %then %do;
        if &eventFilter
          %if &eventRKsNoFilter1 ne %str() %then %do;
             or (&eventRKColumnName in (
               %do i = 1 %to &numEventRKsNoFilter;
                 &&eventRKsNoFilter&i
               %end;
             ))
          %end;

        then output;
      %end;
	  
    run;
    proc sort data=&pam_fslibref..&filteredSubsetTableName;
      by loc_cd asset_id datetime;
    run;
  %end;
%end;
%mend;

%macro pam_gen_fs_get_fs_attrs;
/* Begin the query. The DATA statement names the output data set.  */
data work.metadata_libraries;

  /* The LENGTH statement defines variables for function arguments and */
  /* assigns the maximum length of each variable.                      */
  length liburi upasnuri $256 name $128 type id $17 libref engine $8 path mdschemaname schema $256;

  /* The KEEP statement defines the variables to include in the output data set.  */
  keep name libref engine path;

  /* The CALL MISSING routine initializes the output variables to missing values.  */
  call missing(liburi,upasnuri,name,engine,libref);

  /* The METADATA_GETNOBJ function specifies to get the SASLibrary objects            */
  /* in the repository. The argument nlibobj=1 specifies to get the first object that */
  /* matches the requested URI. liburi is an output variable. It will store the URI   */
  /* of the returned SASLibrary object.                                               */
  nlibobj=1;
  librc=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",nlibobj,liburi);

  /* Initialize found flag.  Flag indicates when PAMFS library information has been found. */
  found = 0;

  /* The DO statement specifies a group of statements to be executed as a unit           */
  /* for each object that is returned by METADATA_GETNOBJ. The METADATA_GETATTR function */
  /* is used to retrieve the values of the Name, Engine, and Libref attributes of        */
  /* the SASLibrary object.                                                              */

  do while (librc>0 and found=0);
    /* Get Library attributes */
    rc=metadata_getattr(liburi,'Name',name);
    rc=metadata_getattr(liburi,'Engine',engine);
    rc=metadata_getattr(liburi,'Libref',libref);

    if kupcase(libref) eq kupcase("&pam_fslibref") then do;
 
      /* The PAMFS library has been found.  Set flag to exit loop after current iteration. */
      found = 1;

      /* The METADATA_GETNASN function specifies to get objects associated to the           */
      /* library via the UsingPackages association. The n argument specifies to return the  */
      /* first associated object for that association type. upasnuri is an output variable. */
      /* It will store the URI of the associated metadata object, if one is found.          */
      n=1;
      uprc=metadata_getnasn(liburi,'UsingPackages',n,upasnuri);

      /* When a UsingPackages association is found, the METADATA_RESOLVE function           */
      /* is called to resolve the URI to an object on the metadata server. The CALL MISSING */
      /* routine assigns missing values to output variables.                                */
      if uprc > 0 then do;
        call missing(type,id,path,mdschemaname,schema);
        rc=metadata_resolve(upasnuri,type,id);

        /* If type='Directory', the METADATA_GETATTR function is used to get its */
        /* path and output the record                                            */
        if type='Directory' then do;
          rc=metadata_getattr(upasnuri,'DirectoryName',path);
          if kupcase(libref) eq kupcase("&pam_fslibref") then do;
            output;
          end;
        end; 

        /* If type='DatabaseSchema', the METADATA_GETATTR function is used to get */ 
        /* the name and schema, and output the record                             */
        else if type='DatabaseSchema' then do;
          rc=metadata_getattr(upasnuri,'Name',mdschemaname);
          rc=metadata_getattr(upasnuri,'SchemaName',schema);
          if kupcase(libref) eq kupcase("&pam_fslibref") then do;
            output;
          end;
        end; 

        /* Check to see if there are any more Directory objects */
        n+1;
        uprc=metadata_getnasn(liburi,'UsingPackages',n,upasnuri);
      end; /* if uprc > 0 */
    end; /* if libref = pamfs */

    /* Look for another library */
    nlibobj+1;
    librc=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",nlibobj,liburi);
  end; /* do while (librc>0) and found = 0 */
run;
%if &sqlobs eq 0 %then %do;
  /* PAMFS library information was not found in the metadata server. */
  %let pam_fs_size = .;
%end;
%else %do;
  proc sql noprint;
    select distinct path into :fsPath
    from work.metadata_libraries
    where kupcase(libref) eq kupcase("&pam_fslibref");
  quit;
  
  data _null_;
    %if &sysscp eq WIN %then %do;
      file='"'||ktrim(kleft(%quote(tranwrd("&fsPath","/","\"))))||%str("\")||ktrim(kleft("&data_sel_id"))||".sas7bdat"||'"';
    %end;
    %else %if &sysscp eq LINUX or &sysscp eq %str(LIN X64) %then %do;
      file='"'||ktrim(kleft(%quote(tranwrd("&fsPath","\","/"))))||%str("/")||klowcase(ktrim(kleft("&data_sel_id")))||".sas7bdat"||'"';
    %end;

    rc=filename("physFile",%quote(file));
    fid=fopen("physFile");
    %if &sysscp eq WIN %then %do;
      filesizeKB=input(finfo(fid,foptname(fid,4)),32.)/1024;
    %end;
    %else %if &sysscp eq LINUX or &sysscp eq %str(LIN X64) %then %do;
      filesizeKB=input(finfo(fid,foptname(fid,6)),32.)/1024;
    %end;

    call symputx('pam_fs_size', fileSizeKB);
    close=fclose(fid);
  run;

%end;  /* End if PAMFS library was found in metadata */
  
%if %sysfunc(exist(&pam_fslibref..&data_sel_id)) %then %do;
  data _null_;
    dsid = open("&pam_fslibref..&data_sel_id");
    if (dsid eq 0) then do; 
      return;
    end;
    else do;
      call symputx('pam_fs_num_cols', attrn(dsid, "NVARS"));
      call symputx('pam_fs_num_obs', attrn(dsid, "NOBS"));
    end;
    dsid = close(dsid);
  run;
%end;

%finish:

%mend;

%macro pam_gen_fs_clr_fmts;
  %let pamformats=;
  /* delete temporary formats if they exist */
  %if %sysfunc(cexist(&pam_fmtlibref..formats.assetrkfmt_&data_sel_id._Z.format)) %then 
    %let pamformats = &pamformats assetrkfmt_&data_sel_id._Z.format;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.eventrkfmt.format)) %then 
    %let pamformats = &pamformats eventrkfmt.format;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.locfmt_&data_sel_id._Z.format)) %then 
    %let pamformats = &pamformats locfmt_&data_sel_id._Z.format;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.measurerkfmt.format)) %then 
    %let pamformats = &pamformats measurerkfmt.format;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.tagtoassetidfmt_&data_sel_id._Z.format)) %then 
    %let pamformats = &pamformats tagtoassetidfmt_&data_sel_id._Z.format;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.tagtoassetrkfmt_&data_sel_id._Z.format)) %then 
    %let pamformats = &pamformats tagtoassetrkfmt_&data_sel_id._Z.format;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.eventidtorkfmt.formatc)) %then 
    %let pamformats = &pamformats eventidtorkfmt.formatc;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.idtonamefmt.formatc)) %then 
    %let pamformats = &pamformats idtonamefmt.formatc;
  %if %sysfunc(cexist(&pam_fmtlibref..formats.tagidtorkfmt.formatc)) %then 
    %let pamformats = &pamformats tagidtorkfmt.formatc;

  %if %sysfunc(cexist(&pam_fmtlibref..formats)) and &pamformats ne  %then %do;
    proc catalog cat=&pam_fmtlibref..formats;
      delete &pamformats;
    quit;
  %end;
%mend;

%macro pam_gen_fs(data_selection_id=);
 /*+----------------------------------------------------------------------+
   |
   | Name:     pam_gen_fs macro
   |
   | Purpose:  This is the main section of the Data Selection macro
   |           code.  There are several macros involved in the
   |           creation of a filtered subset.
   |
   |
   | Product:  SAS Asset Performance Analytics
   | Support:  Cheryl Davis
   |
   | Parms:    Data Selection ID
   |
   | Output:   Filtered Subset
   |
   | History:  date    action
   |           30Sep11 Initial coding                               cld
   |           23Aug12 Support events data - untransposed           cld
   |           04Sep12 Support events data - transposed             cld
   |           04Sep12 Integrate with major changes in 
   |                   pam_expand_transpose macro for 5.2           cld 
   |           04Sep12 Read mart metadata from new pamdo tables in
   |                   Manufacturing Commons structure              cld
   |           01Oct12 Add support for filtering of events data     cld
   |           16Oct12 Remove dictionary table references           cld
   |
   | Notes:    None
   |
   | Copyright (c) 2012 by SAS Institute Inc., Cary, NC, USA.
   | All Rights Reserved.
   |
   +----------------------------------------------------------------------+*/

%pam_prebuild;

%global filteredSubsetTableName;
%global measure_val_col_nm;
%global measure_dttm_col_nm;
%global event_value_col_nm;
%global event_dttm_col_nm;
%global tagDimTable;
%global eventDimTable;
%global assetDimTable;
%global locDimTable;
%global tagRKList;
%global eventRKList;
%global assetRKList;
%global p_table_names;
%global RKcolumnName;
%global eventRKcolumnName;
%global tagIdColumnName;
%global eventIdColumnName;
%global tagTransposeColumnName;
%global eventTransposeColumnName;
%global transposeFlag;
%global startDateTag;
%global endDateTag;
%global startDateEvent;
%global endDateEvent;
%global startNumTag;
%global endNumTag;
%global startNumEvent;
%global endNumEvent;
%global startTimeSlice;
%global endTimeSlice;
%global startDatetime;
%global endDatetime;
%global pam_fs_num_cols;
%global pam_fs_num_obs;
%global pam_fs_size;
%global pam_gen_fs_rc;
%global extractEvents;
%global extractMeasures;
%global interpolationMethodTag;
%global aggregationMethodTag;
%global aggregationMethodEvent;
%global periodicitySeconds;
%global contentTypeCodes;
%global engine;
%global partitionsTag;
%global partitionsEvent;

%local measureFactTable;
%local eventFactTable;

%let startDateTag=;
%let endDateTag=;
%let startDateEvent=;
%let endDateEvent=;
%let startNumTag=;
%let endNumTag=;
%let startNumEvent=;
%let endNumEvent=;
%let startTimeSlice=;
%let endTimeSlice=;
%let startDatetime=;
%let endDatetime=;
%let filteredSubsetTableName=;
%let measure_val_col_nm=;
%let measure_dttm_col_nm=;
%let event_value_col_nm=;
%let event_dttm_col_nm=;
%let tagDimTable=pam_tag_dim;
%let eventDimTable=pam_event_dim;
%let assetDimTable=pam_asset_dim;
%let locDimTable=pam_loc_dim;
%let tagRKList=;
%let eventRKList=;
%let assetRKList=;
%let p_table_names=;
%let RKcolumnName=;
%let eventRKcolumnName=;
%let tagIdColumnName=;
%let eventIdColumnName=;
%let tagTransposeColumnName=;
%let eventTransposeColumnName=;
%let transposeFlag=;
%let pam_fs_num_cols=;
%let pam_fs_num_obs =;
%let pam_fs_size=;
%let pam_gen_fs_rc = 0;
%let extractEvents = N;
%let extractMeasures = N;
%let interpolationMethodTag=;
%let aggregationMethodTag=;
%let aggregationMethodEvent=;
%let periodicitySeconds=3600;
%let engine=;
%let partitionsTag=;
%let partitionsEvent=;

data &pam_dolibref..data_sel;
  modify &pam_dolibref..data_sel;
  if data_sel_id eq kupcase("&data_selection_id") or data_sel_id eq kupcase("&p_data_selection_id");
  run_dttm = datetime();
  data_sel_status_cd = "RUNNING";
run;

%if &p_data_selection_id ne "" %then %let data_sel_id = %sysfunc(kupcase(&p_data_selection_id));
%else %if &data_selection_id ne "" %then %let data_sel_id = %sysfunc(kupcase(&data_selection_id));
%else %do;
  %put ERROR: Data selection ID is required.;
  %let pam_gen_fs_rc = 8;
  %goto finish;
%end;

%pam_gen_fs_clr_fmts;

proc sql noprint;
  select data_sel_id
  from &pam_dolibref..data_sel
  where data_sel_id eq "&data_sel_id";
quit;

%if &sqlobs eq 0 %then %do;
  %put ERROR: No data selection definition exists for selection ID &data_sel_id.;
  %let pam_gen_fs_rc = 8;
  %goto finish;
%end;

data _null_;
  set &pam_dolibref..data_sel_fact;
  if data_sel_id eq "&data_sel_id";

  if upcase(fact_tbl_cd) eq "TAGS" then do;
    call symput('extractMeasures',"Y");
  end;
  if upcase(fact_tbl_cd) eq "EVENTS" then do;
    call symput('extractEvents',"Y");
  end;
run;

%pam_gen_fs_get_dim_table_names;

proc sql noprint;
  create table work.locations as
  select distinct loc_cd, pam_loc_dim_rk
  from &pam_dimlibref..&locDimTable as pam_loc_dim
  where current_num eq 1;
quit;
data locfmt(keep=fmtname start label type hlo);
  set work.locations end = eof;
  retain fmtname "locfmt_&data_sel_id._Z" type "N";
  start=pam_loc_dim_rk;
  label=loc_cd;
  output;
  if (eof) then 
  do;
     hlo = "O";
     start = -999;
     label="UNKNOWN";
     output;
  end;
run;
proc sort data=work.locfmt;
  by start label;
run;
proc format library=&pam_fmtlibref cntlin=locfmt;
run;

%if &extractMeasures eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_mart_meta_tag;
%end;
%if &extractEvents eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_mart_meta_event;
%end;

%pam_gen_fs_get_data_sel(data_selection=&data_sel_id);

data _null_;
  set &pam_dolibref..data_sel_content(where=(data_sel_id eq "&data_sel_id"));
  length temp $100;
  retain temp;
  retain foundTags foundEvents 0;

  if not foundTags and kupcase(content_type_cd) eq "MAINSENSORSOURCE" then do;
    temp=ktrim(kleft(temp))||" TAGS";
    foundTags = 1;
  end;
  else if not foundEvents and kupcase(content_type_cd) eq "MAINEVENTSOURCE" then do;
    temp=ktrim(kleft(temp))||" EVENTS";
    foundEvents = 1;
  end;
  call symput('contentTypeCodes',temp);
run;

/* Determine the engine for PAMFACT */
%let engine = %pam_get_engine(&pam_factqlibref); 
%let engine = %kupcase(&engine);

%pam_gen_fs_get_content(contentTypes=&contentTypeCodes, dataSelection=&data_sel_id);
%pam_gen_fs_get_filter_content(filterContentTypes=&contentTypeCodes, dataSelection=&data_sel_id);

%if &extractMeasures eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_list_of_tags(data_selection=&data_sel_id);
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_tag_RKs;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_extra_tag_obs;
%end;

%if &extractEvents eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_list_of_assets(data_selection=&data_sel_id);
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_asset_RKs;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_list_of_events(data_selection=&data_sel_id);
%end;

%let startNumTag = %sysfunc(inputn(&startDateTag, date, 9));
%let endNumTag = %sysfunc(inputn(&endDateTag, date, 9));
%let startNumEvent = %sysfunc(inputn(&startDateEvent, date, 9));
%let endNumEvent = %sysfunc(inputn(&endDateEvent, date, 9));

/* ------------------------------------------------------------------- */
/* If the fact tables are in Hadoop, generate a list of all partitions */
/* between the start date and end date.  The list of partitions will   */
/* be added to the where clause when extracting from the fact tables.  */
/* ------------------------------------------------------------------- */
%if &engine eq HADOOP or &engine eq SASIOIMP or &engine eq IMPALA %then %do;
  proc sql noprint; 
    select distinct quote(put(year_no,4.)||put(month_no,z2.)) 
    into :partitionsTag separated by "," 
    from &pam_dimlibref..pam_date_dim 
    where pam_date_dim_rk between &startNumTag and &endNumTag; 
  quit;
  %if (&startDateTag ne &startDateEvent) or (&endDateTag ne &endDateEvent) %then %do;
    proc sql noprint; 
      select distinct quote(put(year_no,4.)||put(month_no,z2.)) 
      into :partitionsEvent separated by "," 
      from &pam_dimlibref..pam_date_dim 
      where pam_date_dim_rk between &startNumEvent and &endNumEvent; 
    quit;
  %end;
  %else %do;
    %let partitionsEvent = &partitionsTag;
  %end;
%end;

%if &extractMeasures eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_measure_facts;
%end;

%if &extractEvents eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_event_facts;
%end;

%if &pam_gen_fs_rc le 4 %then %pam_gen_fs_transpose(data_selection=&data_sel_id);

%finish:

%pam_gen_fs_get_fs_attrs;
%pam_gen_fs_clr_fmts;

/* Update PAMDO.DATA_SEL table with filtered subset attributes. */
data &pam_dolibref..data_sel;
  modify &pam_dolibref..data_sel;
  if data_sel_id eq "&data_sel_id";
  %if &pam_gen_fs_rc le 7 %then %do;
    filtered_subset_rows_no = &pam_fs_num_obs;
    filtered_subset_columns_no = &pam_fs_num_cols;
    filtered_subset_size_no = &pam_fs_size;
    data_sel_status_cd = "SUCCESS";
  %end;
  %else %do;
    data_sel_status_cd = "FAILURE";
  %end;
run;

%if &transposeFlag eq BLANK %then %do;
  %put **** WARNING **** WARNING **** WARNING **** WARNING **** WARNING **** WARNING **** ;
  %put;
  %put WARNING: DATA_SEL.DATA_SEL_TRANSPOSE_FLG contains a missing value for ;
  %put %str(        ) this data selection. ;
  %put;
  %put %str(        ) The resulting filtered subset %sysfunc(upcase(&pam_fslibref)).&data_sel_id will contain ;
  %put %str(        ) non-transposed data, regardless of the selection that was ;
  %put %str(        ) made for the transpose option in the Data Selection Wizard. ;
  %put;
  %put ***********************************************************************************;
  %put;   %put;   %put;
%end;

%put ******************** APA Data Selection Report *********************;
%put *                                                           ;
%put * %str(  )Filtered Subset: %sysfunc(upcase(&pam_fslibref)).&data_sel_id ;
%put *                                                           ;
%if &pam_gen_fs_rc gt 7 %then %do;
  %put * %str(  )Data selection status:          Failure         ;
%end;
%else %if &pam_gen_fs_rc ge 4 %then %do;
  %put * %str(  )Data selection status:          Completed with Warnings ;
  %put * %str(  )Number of variables:           &pam_fs_num_cols ;
  %put * %str(  )Number of observations:        &pam_fs_num_obs  ;
  %put * %str(  )Size of filtered subset (KB):  &pam_fs_size     ;
%end;
%else %do;
  %put * %str(  )Data selection status:          Success         ;
  %put * %str(  )Number of variables:           &pam_fs_num_cols ;
  %put * %str(  )Number of observations:        &pam_fs_num_obs  ;
  %put * %str(  )Size of filtered subset (KB):  &pam_fs_size     ;
%end;
%put *                                                           ;
%put *********************************************************************;

%mend;
